import click
import os
import pytz
import shlex
import sys
import threading
import time
import traceback

from tabulate import tabulate
from websocket._exceptions \
  import WebSocketConnectionClosedException \
  as WebSocketClosedError

from . import clicker
from . import bits
from .bits import out, err, dbg, attrs, loads, dumps, spin
from .service import RustService

rc = attrs(
  address=None,
  port=28016,
  password=None,
  connect_timeout=15,
  response_timeout=5,
  console_log_count=45,
  chat_log_count=45,
  debug=1,
  dump=False,
  trace=False,
  trace_filter=1,
  service=None,
  rc_path=None,
  rc_paths=(
    os.path.join(os.path.expanduser("~"), ".config", "rucorc"),
    os.path.join(os.path.expanduser("~"), ".rucorc"),
    "rucorc",
    ".rucorc",
  ),
)

cli_spec = """

name: ruco

help: |
  Interact with a Rust game server admin console

options:

  - name: -a --address
    help: Server IP address or host name
    type: str
    required: yes

  - name: --port
    help: Server port, default 28016
    type: int
    default: {port}

  - name: -p --password
    help: Server password; use 'ask' to enter via a prompt
    type: str
    required: yes

  - name: -c --connect-timeout
    help: Connection timeout, default {connect_timeout}
    type: int
    default: {connect_timeout}

  - name: -r --response-timeout
    help: Command response timeout, default {response_timeout}
    type: int
    default: {response_timeout}

  - name: --debug
    help: "0: error and warning, 1: debug, 2: spam, default {debug}"
    type: int
    default: {debug}

  - name: --dump
    help: Dump messages
    is_flag: yes

  - name: --trace
    help: Trace function/method calls
    is_flag: yes

  - name: --trace-filter
    help: "0: trace all calls, 1: trace ruco calls only, default: {trace_filter}"
    type: int
    default: {trace_filter}

commands:

  - name: rcon
    help: Send a command to the server

    options:

      - name: -q --quiet
        help: Don't wait for a response after sending the command
        is_flag: yes

    arguments:

      - name: command
        nargs: -1
        required: yes

  - name: players
    help: Show connected players

  - name: console
    help: Print console log

    options:

      - name: -f --follow
        help: Follow console log until interrupted
        is_flag: yes

      - name: -n --count
        help: Show a number of lines from the end, default {console_log_count}
        type: int
        default: {console_log_count}

  - name: chat
    help: Print chat log

    options:

      - name: -f --follow
        help: Follow chat log until interrupted
        is_flag: yes

      - name: -n --count
        help: Scan a number of lines from the end for chat messages, default {chat_log_count}
        type: int
        default: {chat_log_count}

  - name: say
    help: Send a chat message to players

    arguments:

      - name: message
        nargs: -1
        required: yes

""".format(**rc)

class RucoException(Exception): pass
class ConnectTimeoutError(RucoException): pass
class ResponseTimeoutError(RucoException): pass

def exit(status=0):
  s = rc.service
  if s:
    on_error in s.on_error and s.on_error.remove(on_error)
    s.connected and s.disconnect()
  # This function could be called by threads other than main, so use os._exit
  os._exit(status)
  #sys.exit(status)

def error(exc_info):
  if isinstance(exc_info, Exception):
    exc_info = bits.make_exc_info(exc_info)
  if rc.debug:
    dbg("".join(traceback.format_exception(*exc_info)))
  else:
    err("%s: %s" % (exc_info[1].__class__.__name__, exc_info[1]))
  exit(-1)

def on_error(svc, exc_info):
  error(exc_info)

# FIXME I'm tired and this is awful:

def start_death_clock(name, timeout, extype=None):
  def on_timeout():
    threading.currentThread().setName("<%s>" % name)
    bits.trace(enabled=rc.trace, filt=rc.trace_filter)
    extype and error(extype("Timed out")) or exit()
  t = getattr(start_death_clock, "timer", None)
  if t:
    t.cancel()
  t = start_death_clock.timer = threading.Timer(timeout, on_timeout)
  t.start()

def stop_timer():
  t = getattr(start_death_clock, "timer", None)
  if t:
    t.cancel()
    start_death_clock.timer = None

def start_connect_timer(extype=ConnectTimeoutError):
  start_death_clock("[connect timer]", rc.connect_timeout, extype)

def start_response_timer(extype=ResponseTimeoutError):
  start_death_clock("[response timer]", rc.response_timeout, extype)

# FIXME End awful

def call(f, *args, **kwargs):
  try:
    return f(*args, **kwargs)
  except SystemExit:
    raise
  except KeyboardInterrupt:
    exit()
  except:
    error(sys.exc_info())

def connect(on_connect):
  def on_connect_cancel_timer(svc):
    stop_timer()
  s = rc.service = RustService(
    rc.address,
    rc.port,
    rc.password,
    dump=rc.dump
  )
  s.on_connect.append(on_connect_cancel_timer)
  s.on_connect.append(on_connect)
  start_connect_timer()
  s.connect()

def display(msg):
  if rc.dump:
    # The service will dump all messages in dump mode
    return
  stamp = time.strftime("%H:%M:%S", time.localtime(msg.get("Time")))
  type = "Type" in msg and msg.Type and "[%s] " % msg.Type or ""
  text = "%s %s%s" % (stamp, type, msg.Message)
  if "Username" in msg:
    text = "%s <%s> %s" % (stamp, msg.Username, msg.Message)
  else:
    try:
      text = "%s %s" % (stamp, dumps(loads(msg.Message), indent=4))
    except:
      pass
  out(text)

def ruco(
  address,
  port,
  password,
  connect_timeout,
  response_timeout,
  debug,
  dump,
  trace,
  trace_filter
):
  def _ruco():
    rc.address = address
    rc.port = port
    if password is "ask":
      raise Exception("not implemented") # FIXME
    rc.password = password
    nonlocal debug
    rc.debug = debug in (0, 1, 2) and debug or 0
    rc.dump = dump
    rc.trace = trace
    rc.trace_filter = trace_filter in (0, 1) and trace_filter or 1
    bits.trace(enabled=trace, filt=trace_filter)
  call(_ruco)

def ruco_rcon(quiet, command):
  timer_reset = attrs(value=False)
  def on_message_send(svc, msg):
    exit()
  def on_command_response(svc, msg):
    display(msg)
    if timer_reset.value is False:
      start_response_timer(extype=None)
      timer_reset.value = True
  def on_connect(svc):
    if quiet:
      svc.on_message_send.append(on_message_send)
    else:
      start_response_timer()
    svc.request(" ".join(command), on_command_response)
  call(connect, on_connect)

def ruco_players():
  def on_playerlist_response(svc, msg):
    players = loads(msg.Message)
    if not players:
      out("No players connected.")
      exit()
    headers = [
      "Name",
      "Steam ID",
      "Ping",
      "Address",
      "Connected",
      "Health",
      "Violation",
    ]
    players = (
      (
        player.DisplayName,
        player.SteamID,
        player.Ping,
        player.Address,
        bits.format_seconds(player.ConnectedSeconds),
        player.Health,
        player.VoiationLevel,
      ) for player in players
    )
    out(tabulate(players, headers=headers))
    exit()
  def on_connect(svc):
    svc.request("playerlist", on_playerlist_response)
  call(connect, on_connect)

def _ruco_log(follow, count, cmd, filt=None):
  def on_message_recv(svc, msg):
    if not filt or not filt(msg):
      display(msg)
  def on_tail_response(svc, msg):
    logs = loads(msg.Message)
    if len(logs) == 0:
      print("No messages available.")
    else:
      for log in logs:
        display(log)
    if follow:
      svc.on_message_recv.append(on_message_recv)
    else:
      exit()
  def on_connect(svc):
    svc.request("%s %d" % (cmd, count), on_tail_response)
  call(connect, on_connect)

def ruco_console(follow, count):
  _ruco_log(follow, count, "console.tail")

def ruco_chat(follow, count):
  filt = lambda m: "Type" in m and m.Type != "Chat"
  _ruco_log(follow, count, "chat.tail", filt)

def ruco_say(message):
  def on_message_send(svc, msg):
    exit()
  def on_connect(svc):
    svc.on_message_send.append(on_message_send)
    svc.command("say %s" % " ".join(message))
  call(connect, on_connect)

def get_rc_path():
  envrc = "RUCO_RC" in os.environ and os.environ["RUCO_RC"] or None
  paths = envrc and (envrc,) + rc.rc_paths or rc.rc_paths
  for p in paths:
    if os.path.isfile(p):
      return p
  return None

def read_rc(path):
  with open(path) as rc:
    return dict(token.split("=") for token in shlex.split(rc.read()))

def load_rc():
  rc.rc_path = p = get_rc_path()
  if not p:
    return
  try:
    for k, v in read_rc(p).items():
      if k not in os.environ:
        os.environ[k] = v
  except:
    err("%s%sError loading config file: %s" % (
      traceback.format_exc(), os.linesep, p
    ))
    sys.exit(-1)

def main():
  threading.current_thread().setName("<main>")
  load_rc()
  # Start tracing as early as possible when set via the config file or an
  # environment variable
  bits.trace(enabled=rc.trace, filt=rc.trace_filter)
  cli = clicker.Cli()
  cli.loads(cli_spec)
  cli.run(env=globals(), require_groups=True)

if __name__ == "__main__":
  main()



import collections
import inspect
import json
import os
import sys
import time
import threading
import traceback

DEBUG = 3

def format_seconds(i):
  h, r = divmod(i, 3600)
  m, s = divmod(r, 60)
  return "%sh%sm%ss" % (h, m, s)

def xttrs(type, *args, **kwargs):
  """
  Wrap an instance of a mapping type such that keys may be accessed as
  attributes. The mapping super-type may be specified, defaulting to dict.
  """
  type = type and type or dict
  class _(type):
    __setattr__ = type.__setitem__
    __delattr__ = type.__delitem__
    def __init__(self):
      super().__init__(*args, **kwargs)
    def __getattr__(self, name):
      if name not in self:
        raise AttributeError("No such attribute: %s" % name)
      return self.__getitem__(name)
  return _()

def attrs(*args, **kwargs):
  return xttrs(dict, *args, **kwargs)

def ottrs(*args, **kwargs):
  return xttrs(collections.OrderedDict, *args, **kwargs)

fds = attrs(
  out=sys.stdout,
  err=sys.stderr,
)

def out(*args):
  print(*args, file=fds.out)

def err(*args):
  print(*args, file=fds.err)

def warn(*args):
  DEBUG > 0 and print("WARN", *args, file=fds.err)

def dbg(*args):
  DEBUG > 1 and print("DBG", *args, file=fds.err)

def spam(*args):
  DEBUG > 2 and print("SPAM", *args, file=fds.err)

def dispatch(handlers, *args, **kwargs):
  errors = []
  for fn in handlers:
    try:
      fn(*args, **kwargs)
      errors.append(None)
    except:
      errors.append(sys.exc_info())
      dbg(traceback.format_exc())
      dbg("%s%sUnhandled exception in event dispatch" % (
        os.linesep, traceback.format_exc()
      ))
  return errors

def loads(s, *a, **kw):
  if hasattr(s, "decode"):
    s = s.decode("utf-8")
  return json.loads(s, *a, object_pairs_hook=ottrs, **kw)

dumps = json.dumps

def make_exc_info(ex):
  if isinstance(ex, SystemExit):
    raise ex
  try:
    raise ex
  except:
    return sys.exc_info()

def spin(timeout, interval, test):
  if interval < 0:
    raise ValueError("'interval' must be >= 0")
  if timeout > 0 and interval > 0:
    interval = min(timeout, interval)
  remaining = timeout > 0 and timeout or 1
  while remaining > 0:
    start = time.time()
    if test and test():
      return True
    if timeout == 0:
      return False
    test_time = time.time() - start
    sleep_time = interval - test_time
    if sleep_time < 0:
      sleep_time = 0
    if timeout < 0:
      time.sleep(sleep_time)
    else:
      start = time.time()
      remaining -= test_time
      if remaining > 0:
        time.sleep(min(sleep_time, remaining))
        remaining -= time.time() - start

def trace(enabled=True, filt=True, indent=True, ignore_self=True):
  def f_module(frame):
    return (
      "__name__" in frame.f_globals and
      "%s." % frame.f_globals["__name__"] or
      ""
    )
  def f_class(frame):
    args, _, _, values = inspect.getargvalues(frame)
    if args and values and "self" in values:
      obj = values["self"]
      return (
        hasattr(obj, "__class__") and
        hasattr(obj.__class__, "__name__") and
        "%s." % obj.__class__.__name__ or
        ""
      )
    return ""
  def f_qualname(frame):
    fn = frame.f_code.co_name
    g = frame.f_globals
    if fn in g:
      fn_obj = g[fn]
      if hasattr(fn_obj, "__qualname__"):
        fn = fn_obj.__qualname__
    return fn
  def filter(module): # FIXME DX
    return (
      filt and not (
        not module or
        module.startswith("ruco") or
        module.endswith("ruco") or
        module.startswith("__")
      ))
  def t(frame, event, arg):
    try:
      label = {"call": "CALL", "return": "EXIT"}.get(event)
      if label is None:
        return
      if ignore_self and frame.f_code in (trace.__code__, t.__code__):
        return
      module = f_module(frame)
      if filter(module):
        return
      sys.stderr.write(
        "%s %s %s %s%s%s%s" % (
          threading.currentThread().name,
          indent and "." * len(inspect.stack()) or "",
          label,
          module,
          f_class(frame),
          f_qualname(frame),
          os.linesep
        ))
    except:
      pass
  if enabled:
    sys.setprofile(None)
    sys.setprofile(t)
  else:
    sys.setprofile(None)



"""
clicker - rapid command-line user interface development

- Provides convenient syntax and semantics for constructing command-line
  interfaces definitions, and tools to speed up development of command-line
  applications.

- Define all commands, options, and arguments accepted by an application using
  a straight-forward syntax in yaml or json.

- For simple applications, an argument parser is easily instantiated for a
  CLI definition, and callbacks for commands/options/arguments are
  automatically mapped to Python functions implemented by the user.
  (See the main function of this script for an example of this idiom.)

- For complex applications, skeleton Python source code can be generated for
  command/option/argument handlers from a CLI definition in yaml/json, which
  can then be implemented incrementally by the user.

- The command-line interface definition semantics allow 'inheritance', that
  is, deriving a new CLI definition from an existing one, which could be useful
  for complex applications with many commands that are more similar than
  different.

- Last but far from least, clicker is built using the (outstanding, fantastic,
  amazing, where-would-I-be-without-it) Click toolkit:

    http://click.pocoo.org/

Nicholas A. Zigarovich <nzigarovich@ritchiebros.com>
"""

from __future__ import print_function

import click
import collections
import copy
import json
import sys
import traceback
import yaml

try:
  import IPython
  pp = IPython.lib.pretty.pprint
  def debug():
    traceback.print_exc()
    IPython.embed()
except ImportError:
  pp = print
  import pdb
  def debug():
    traceback.print_exc()
    pdb.pm()

def popkey(d, key, default=None):
  if key in d:
    r = d[key]
    del d[key]
    return r
  return default

def merge(old, new):
  def shift(k):
    if k in new:
      old[k] = new[k]
  shift("name")
  shift("help")
  shift("options")
  shift("arguments")
  if "commands" in new:
    if "commands" not in old:
      old["commands"] = new["commands"]
    else:
      for new_command in new["commands"]:
        try:
          old_command = [
            x for x in old["commands"]
            if x["name"] == new_command["name"]
          ][0]
          old["commands"].remove(old_command)
        except IndexError:
          pass
        old["commands"].append(new_command)
  if "groups" in new:
    if "groups" not in old:
      old["groups"] = new["groups"]
    else:
        for new_group in new["groups"]:
          try:
            old_group = [
              x for x in old["groups"]
              if x["name"] == new_group["name"]
            ][0]
            merge(old_group, new_group)
          except IndexError:
            old["groups"].append(new_group)
  return old

def stub(
  data, fd=sys.stdout, groups=False, get_cb=None, tab="  ", indent=0,
  imports=True
):
  def tabs(): return indent * tab
  def push(): tabs += 1
  def pop(): tabs -= 1
  def p(s): fd.write(s)

  path = []

  if get_cb is None:
    get_cb = lambda p: "_".join(p)

  def build_options(o):
    pass

  def print_command(c):
    paths.append(c["name"])
    p(tabs() + "def %s(%s):\n" % (get_cb(path), build_options(c)))
    push()
    p(tabs() + "pass\n\n")
    pop()
    paths.pop()

  def print_commands(g):
    for c in g.get("commands", ()):
      print_command(c)

  def print_group(g):
    paths.append(g["name"])
    if groups:
      p(tabs() + "def %s(%s):\n" % (get_cb(path), build_options(g)))
      push()
      p(tabs() + "pass\n\n")
      pop()
    print_commands(g)
    paths.pop()

  def print_groups(g):
    for gg in g.get("groups", ()):
      print_group(gg)

  if imports:
    p(tabs() + "import click\n\n")
    p(tabs() + "get_context = click.get_current_context\n")
    p(tabs() + "get_obj = lambda: get_context().obj\n\n")
  print_group(data)

def build(
  data, env=None, get_cb=None, require_commands=True, require_groups=False
):
  path = []

  if get_cb is None:

    def get_cb(p, r):
      n = "_".join(p)
      f = (env or globals()).get(n)
      if not f and r:
        raise KeyError("Required callback not found in globals(): %s" % n)
      return f

  def build_argument(a):
    a = copy.copy(a)
    name = popkey(a, "name")
    a["type"] = eval(a.get("type", "None"), {"click": click})
    a["default"] = eval(a.get("default", "None"))
    a["nargs"] = eval(a.get("nargs", "None"))
    return click.Argument([name], **a)

  def build_arguments(c):
    return [build_argument(x) for x in c.get("arguments", ())]

  def build_option(o):
    o = copy.copy(o)
    name = popkey(o, "name").split(" ")
    o["type"] = eval(o.get("type", "None"), {"click": click})
    o["default"] = eval(o.get("default", "None"))
    for n in name:
      if n.startswith("--"):
        break
    else:
      n = None
    if n:
      o["envvar"] = "%s_%s" % (
        "_".join(path).upper(),
        n[2:].replace("-", "_").upper()
      )
    return click.Option(name, **o)

  def build_options(o):
    return [build_option(x) for x in o.get("options", ())]

  def build_command(c, require_cb=require_commands, cls=click.Command):
    c = copy.copy(c)
    path.append(c["name"])
    try:
      c["callback"] = get_cb(path, require_cb)
      c["params"] = build_options(c)
      c["params"].extend(build_arguments(c))
      popkey(c, "options")
      popkey(c, "arguments")
      popkey(c, "commands")
      name = popkey(c, "name")
      return cls(name, **c)
    finally:
      path.pop()

  def build_commands(g):
    return [build_command(x) for x in g.get("commands", ())]

  def build_group(g):
    group = build_command(g, require_cb=require_groups, cls=click.Group)
    try:
      path.append(g["name"])
      for subgroup in build_groups(g):
        group.add_command(subgroup, name=subgroup.name)
      for command in build_commands(g):
        group.add_command(command)
      return group
    finally:
      path.pop()

  def build_groups(g):
    return [build_group(x) for x in g.get("groups", ())]

  if len(data.get("groups", ())) == 0 and len(data.get("commands", ())) == 0:
    rv = build_command(data)
  else:
    rv = build_group(data)
  return rv
  #return build_group(data)

def _setup_yaml():
  def representer(dumper, data):
    return dumper.represent_dict(data.items())
  def constructor(loader, node):
    return collections.OrderedDict(loader.construct_pairs(node))
  tag = yaml.resolver.BaseResolver.DEFAULT_MAPPING_TAG
  yaml.add_representer(collections.OrderedDict, representer)
  yaml.add_constructor(tag, constructor)

JSON = "json"
YAML = "yaml"

def loads(s, type=YAML, data={}):
  "Load from string."
  if type == JSON:
      new_data = json.loads(s, object_pairs_hook=collections.OrderedDict)
  elif type == YAML:
    _setup_yaml()
    new_data = yaml.load(s, Loader=yaml.loader.BaseLoader)
  else:
    raise ValueError("Invalid type: %s" % type)
  return merge(data, new_data)

def loadf(f, type=None, data={}):
  "Load from file."
  if type is None:
    if path.lower().endswith("json"):
      type = JSON
    elif path.lower()[-4:] in (".yml", "yaml"):
      type = YAML
    else:
      raise ValueError("Can't determine file type: %s" % f)
  with open(f) as fd:
    return loads(fd.read(), type=type, data=data)

def loadmf(files, type=None, data={}):
  "Load from many files."
  for f in files:
    load(f, type=type, data=data)
  return data

def loadfd(fd, type=YAML, data={}):
  "Load from file descriptor."
  raise NotImplementedError()

def loadmfd(fds, type=YAML, data={}):
  "Load from many file descriptors."
  raise NotImplementedError()

class Cli:

  def __init__(self):
    self.data = {}
    self.cli = None

  def loads(self, s, type=YAML):
    loads(s, type=type, data=self.data)

  def loadf(self, file, type=None):
    loadf(file, type=type, data=self.data)

  def loadmf(self, files, type=None):
    loadmf(files, data=self.data)

  def loadfd(self, fd, type=YAML):
    loadfd(fd, type=type, data=self.data)

  def loadmfd(self, fds, type=YAML):
    loadmfd(fds, type=type, data=self.data)

  def build(self, *args, **kwargs):
    self.cli = build(self.data, *args, **kwargs)

  def run(self, *args, **kwargs):
    self.build(*args, **kwargs)
    self.cli()

  def clear(self):
    self.__init__()

_yaml = """

name: clicker
help: Do things with clicker CLI definitions
commands:
  - name: merge
    help: Merge multiple definition files into one
    options:
      - name: -o --output
        help: Output file, default -
        type: click.File('wb')
        default: '"-"'
      - name: -f --format
        help: Output format, default yaml
        type: click.Choice(["json", "yaml"])
        default: '"yaml"'
    arguments:
      - name: files
        nargs: -1
        required: yes
  - name: stub
    help: Generate Python stubs from defininition files
    options:
      - name: -o --output
        help: Output file, default -
        type: click.File("wb")
        default: '"-"'
      - name: -g --groups
        help: Generate group callbacks
        is_flag: yes
      - name: --no-imports
        help: Don't generate imports
        is_flag: yes
      - name: -t --tab
        help: Tab string, default '"  "'
        default: '"  "'
      - name: -c --click-stubs
        help: Generate Click stubs
        is_flag: yes
    arguments:
      - name: files
        nargs: -1
        required: yes
"""

def clicker_merge(output, format, files):
  d = loadmf(files)
  if format == YAML:
    output.write(yaml.dump(d))
  elif format == JSON:
    output.write(json.dumps(d))

def clicker_stub(output, groups, no_imports, tab, click_stubs, files):
  d = loadmf(files)
  stub(d, fd=output, groups=group, imports=(not no_imports), tab=tab)

def main():
  cli = Cli()
  cli.loads(_yaml)
  cli.run(require_groups=True)

if __name__ == "__main__":
  main()




import os
import sys
import threading
import time
import websocket

from distutils.util import strtobool

from . import bits
from .bits import out, err, dbg, spam, attrs, loads, dumps, dispatch

class RustService(object):

  REQUEST_EXPIRY = 300

  def __init__(self, address, port, password, dump=False):
    super().__init__()
    self.address = address
    self.port = port
    self.password = password
    self.dump = dump
    self.requests = {}
    self.socket = None
    self.on_connect = []
    self.on_disconnect = []
    self.on_message_recv = []
    self.on_message_send = []
    self.on_error = []
    self.on_stale_cleanup = []
    self._identifier = 1001

  @property
  def identifier(self):
    if self._identifier > 0x7FFFFFFF:
      self._identifier = 1001
    id = self._identifier
    self._identifier += 1
    return id

  def _on_message_recv(self, msg):
    if self.dump:
      err("RECV", dumps(msg))
    id = msg.Identifier
    if id > 1000:
      request = self.requests.get(id)
      if request is not None:
        msg.request = request
        if request.callback:
          dispatch((request.callback,), self, msg)
        return
      else:
        spam("Received response for unknown request: %s" % msg)
    dispatch(self.on_message_recv, self, msg)

  def _on_message_send(self, msg):
    dispatch(self.on_message_send, self, msg)

  def _on_connect(self, socket):
    self.socket = socket
    dispatch(self.on_connect, self)

  def _on_error(self, socket, error):
    dispatch(self.on_error, self, bits.make_exc_info(error))
    try:
      # FIXME hmm...
      self.socket.on_close = None
      self.socket.close()
    except SystemExit:
      raise
    except:
      pass
    socket = None
    self.requests.clear()

  def _on_disconnect(self, socket):
    self.socket = None
    self.requests.clear()
    dispatch(self.on_disconnect, self)

  def _send(self, msg):
    if self.dump:
      err("SEND", dumps(msg))
    self.socket.send(dumps(msg))
    self._on_message_send(msg)

  def _expire_requests(self):
    if len(self.requests) == 0:
      return 0
    now = time.time()
    stale = [
      id for id, request in self.requests.items()
      if (now - request.time) > self.REQUEST_EXPIRY
    ]
    for id in stale:
      del self.requests[id]
    dispatch(self.on_stale_cleanup, self, stale)
    spam("Expired %d requests" % len(stale))
    return stale

  @property
  def connected(self):
    return self.socket is not None

  def connect(self):
    def on_message(s, m):
      self._on_message_recv(loads(m))
    def on_pong(*a, **kw):
      self._expire_requests()
    socket = None
    try:
      websocket.enableTrace(strtobool(os.environ.get("DEBUG_WEBSOCKET", "0")))
      socket = websocket.WebSocketApp(
        "ws://%s:%d/%s" % (self.address, self.port, self.password),
        on_open = self._on_connect,
        on_close = self._on_disconnect,
        on_message = on_message,
        on_error = self._on_error,
        on_pong = on_pong
      )
      socket.run_forever()
    except SystemExit:
      raise
    except KeyboardInterrupt:
      self._on_error(socket, sys.exc_info())
      return
    except:
      self._on_error(socket, sys.exc_info())
      raise

  def disconnect(self):
    if self.socket:
      self.socket.close()

  def command(self, msg, id=None):
    if id is None:
      id = -1
    self._send({
      "Identifier": id,
      "Message": msg,
      "Name": "WebRcon",
    })

  def request(self, msg, cb):
    id = self.identifier
    assert(id not in self.requests)
    self.requests[id] = attrs(
      callback=cb,
      time=time.time(),
    )
    try:
      self.command(msg, id)
    except SystemExit:
      raise
    except:
      del self.requests[id]
      raise

class RustServiceThread(RustService):

  def __init__(self, *a, name="[rust]", **kw):
    super().__init__(*a, **kw)
    self.name = name
    self.thread = None
    self.error = None

  def _run(self):
    try:
      super().connect()
    except SystemExit:
      raise
    except:
      self.error = sys.exc_info()
      #self._on_error(self.socket, self.error)
    finally:
      self.thread = None

  def connect(self):
    self.thread = threading.Thread(target=self._run, name=self.name)
    self.thread.start()

  def disconnect(self, wait=0):
    super().disconnect()
    return self.join(wait)

  def join(self, timeout=-1):
    try:
      if timeout == 0:
        return self.connected
      elif timeout < 0:
        self.thread.join()
        return True
      else:
        self.thread.join(timeout)
        return self.connected
    except AttributeError:
      return self.connected



from setuptools import setup

setup(
  name="ruco",
  packages=["ruco"],
  version="0.6",
  description="Rust (by Facepunch Studios) rcon API and shell utility",
  author="exo",
  author_email="exo@eckso.io",
  url="https://github.com/nizig/ruco",
  download_url="https://github.com/nizig/ruco/tarball/release-0.6",
  zip_safe=False,
  entry_points={
    "console_scripts": ["ruco=ruco.cli:main"],
  },
  install_requires=[
    "PyYAML",
    "click",
    "pytz",
    "tabulate",
    "tzlocal",
    "websocket-client",
  ],
  keywords=["rust", "rcon", "game", "console", "facepunch"],
  classifiers=[],
)
